import os
import math
import json
import logging
import hangul_jamo

VOVEL_SPLIT = {
    'ㅘ': 'ㅗㅏ',
    'ㅙ': 'ㅗㅐ',
    'ㅚ': 'ㅗㅣ',
    'ㅝ': 'ㅜㅓ',
    'ㅞ': 'ㅜㅔ',
    'ㅟ': 'ㅜㅣ',
    'ㅢ': 'ㅡㅣ',
}

def decompose(text):
    """
    Regular text to jamo, also split vovels.
    """
    r = hangul_jamo.decompose(text)
    for k, v in VOVEL_SPLIT.items():
        r = r.replace(k, v)
    return r

def compose(text):
    """
    Jamo to regular text, also merge vovels.
    """
    r = hangul_jamo.compose(text)
    for k, v in VOVEL_SPLIT.items():
        r = r.replace(v, k)
    return r

def trace_to_tokens(surface_text, trace):
    """
    Emit tokens from rule from a lemmatisation trace.
    """
    from .inference import Token

    tokens = []

    ta_ending = surface_text.endswith('다')

    offset_map = {}
    accu = 0
    for i, c in enumerate(surface_text):
        dc = decompose(c)
        for ii in range(len(dc)):
            offset_map[accu + ii] = i
        accu += len(dc)
    
    for i, (name, suffix_in, suffix_out, jamo_end_distance) in enumerate(trace):
        is_last = i == len(trace) - 1

        jamo_check_pos = jamo_end_distance - len(suffix_out)
        start_i = len(surface_text) - offset_map[jamo_check_pos] if jamo_check_pos in offset_map else 0
        start_i = max(0, start_i)
        end_i = start_i + math.ceil(len(suffix_in) / 2.4)
        end_i = min(len(surface_text), end_i)
        surface = surface_text[start_i:end_i]

        if name == '-아/어':
            lemma = '아'
            if any(c in suffix_in for c in 'ㅓㅕㅖ'):
                lemma = '어'
            token = Token(surface, lemma, 'EC', start_i, end_i)
            tokens.append(token)
        elif name == '-았/었':
            lemma = '았'
            if any(c in suffix_in for c in 'ㅓㅕㅖ'):
                lemma = '었'
            token = Token(surface, lemma, 'EC', start_i, end_i)
            tokens.append(token)
        elif name == '-[느]ㄴ다':
            lemma = 'ㄴ'
            if suffix_in.startswith('ㄴㅡㄴ'):
                lemma = '는'
            if ta_ending and is_last:
                lemma += '다'
            token = Token(surface, lemma, 'ETM', start_i, end_i)
            tokens.append(token)

    return tokens


class Lemmatizer:
    def __init__(self, data_dir):
        lemmas_path = os.path.join(data_dir, 'lemmas.txt')
        transforms_path = os.path.join(data_dir, 'transforms.json')

        self.lemmas = {}
        with open(lemmas_path, 'r', encoding='utf-8') as f:
            for line in f:
                lemma, freq = line.strip().split()
                self.lemmas[lemma] = int(freq)

        with open(transforms_path, 'r', encoding='utf-8') as f:
            transforms = json.load(f)

        self.rules = []
        for name, transform_rules in transforms.items():
            for transform_rule in transform_rules:
                self.rules.append((name, transform_rule))

    def transform_jamo(self, source_text, max_depth=5):
        """
        Returns a list of all possible lemmatizations of the given text in jamo form.
        """
        results = [(source_text, set(), [], 0)]

        i = 0
        while i < len(results):
            (text, conditions, trace, depth) = results[i]
            if depth >= max_depth:
                i += 1
                continue
            
            # attempt to apply all rules
            for name, (suffix_in, suffix_out, cond_in, cond_out) in self.rules:
                if cond_in and not any(c in conditions for c in cond_in):
                    continue
                if not text.endswith(suffix_in):
                    continue
                
                # exchange the suffix
                new_text = text[:-len(suffix_in)] + suffix_out
                new_conditions = cond_out
                
                old_jamo_end_distance = 0 if not trace else trace[-1][3]
                jamo_end_distance = len(suffix_in) - len(suffix_out) + old_jamo_end_distance
                
                new_trace = trace + [(name, suffix_in, suffix_out, jamo_end_distance)]

                # add new transformation (will be processed later in the loop again)
                results.append((new_text, new_conditions, new_trace, depth + 1))

            i += 1

        return [
            (text, conditions, trace)
            for text, conditions, trace, *_ in results
        ]
    
    def transform(self, source_text):
        """
        Returns a list of all possible lemmatizations of the given text.
        """
        text_jamo = decompose(source_text)
        transformed_jamo = self.transform_jamo(text_jamo)
        transformed = []
        for text_jamo, conditions, trace in transformed_jamo:
            try:
                text = compose(text_jamo)
            except ValueError:
                continue
            transformed.append((text, conditions, trace))
        return transformed

    def lemmatize(self, surface_text):
        """
        Find the best lemmatization of the given text.
        """
        transforms = self.transform(surface_text)
        lemmas = []

        for text, _conditions, trace in transforms:
            # Only consider lemmas that end with verbic dictionary form
            if text.endswith('다'):
                lemma = text[:-1]
                # Only consider lemmas that are in the dictionary
                if not lemma in self.lemmas:
                    continue
                lemmas.append((lemma, trace))
        
        # Fewer transformations are better
        lemmas.sort(key=lambda x: len(x[1]))

        possible_lemmas = []
        for i, (lemma, transforms) in enumerate(lemmas):
            if not lemma in possible_lemmas:
                possible_lemmas.append(lemma)
            logging.debug(f'Lemmatization {i+1}: {lemma} {transforms}')
        logging.debug(f'Possible lemmas: {", ".join(possible_lemmas)}')

        if lemmas:
            lemma, trace = lemmas[0]
            return lemma, trace_to_tokens(surface_text, trace)

        return text, []


def lemmatize(
    data_dir: str,
    **_kwargs,
):
    lemmatizer = Lemmatizer(data_dir)

    while True:
        try:
            text = input('> ')
            print(lemmatizer.lemmatize(text))
        except KeyboardInterrupt:
            break
        except EOFError:
            break



import os
import zipfile
import logging
import requests

entries = {
    'airbnb_review.txt': 'https://github.com/yiunsr/boddari/raw/refs/heads/main/boddari/txt/airbnb_review.txt.zip',
    'book_review01.txt': 'https://github.com/yiunsr/boddari/raw/refs/heads/main/boddari/txt/book_review01.txt',
    'encyclopedia01.01.txt': 'https://github.com/yiunsr/boddari/raw/refs/heads/main/boddari/txt/encyclopedia01.01.txt.zip',
    'fairytale01.txt': 'https://github.com/yiunsr/boddari/raw/refs/heads/main/boddari/txt/fairytale01.txt',
    'policy_news.01.txt': 'https://github.com/yiunsr/boddari/raw/refs/heads/main/boddari/txt/policy_news.01.txt.zip',
    'nsmc.txt': 'https://github.com/e9t/nsmc/raw/refs/heads/master/ratings.txt',
    'unlabeled_comments_1.txt': 'https://github.com/kocohub/korean-hate-speech/raw/refs/heads/master/unlabeled/unlabeled_comments_1.txt',
}

def data_dl(out_dir, **_kwargs):
    """
    Downloads all the data files to the specified directory. Unzips if necessary.
    """
    os.makedirs(out_dir, exist_ok=True)

    for filename, url in entries.items():
        logging.info(f'Downloading {filename} from {url}')

        out_path = os.path.join(out_dir, filename)

        if url.endswith('.zip'):
            out_path = out_path + '.zip'

        with requests.get(url, stream=True) as r:
            r.raise_for_status()
            with open(out_path, 'wb') as f:
                for chunk in r.iter_content(chunk_size=8192):
                    f.write(chunk)

        # Unzip the file if it has a .zip extension
        if out_path.endswith('.zip'):        
            logging.info(f'Unzipping {filename}')
            with zipfile.ZipFile(out_path, 'r') as z:
                z.extractall(out_dir)

            os.remove(out_path)

        # convert nsmc to plain text
        if filename == 'nsmc.txt':
            with open(os.path.join(out_dir, 'nsmc.txt'), 'r', encoding='utf-8') as f:
                lines = f.readlines()
            with open(os.path.join(out_dir, 'nsmc.txt'), 'w', encoding='utf-8') as f:
                for i, line in enumerate(lines):
                    if i == 0:
                        continue
                    line = line.strip()
                    if not line:
                        continue
                    parts = line.split('\t')
                    if len(parts) != 3:
                        continue
                    f.write(parts[1])
                    f.write('\n')

if __name__ == '__main__':
    import sys
    logging.basicConfig(level=logging.INFO, format='[%(levelname)s] %(message)s')
    data_dl(sys.argv[1] if len(sys.argv) > 1 else '.')



import json
import torch
from transformers import AutoConfig, AutoTokenizer, AutoModelForTokenClassification, DataCollatorForTokenClassification, Trainer, TrainingArguments
from . import labels

def train(args):
    # Fitting config for selected model
    config = AutoConfig.from_pretrained(args.model)
    
    # Add classification parameters to config
    config.num_labels = labels.num_labels
    config.id2label = labels.id2label
    config.label2id = labels.label2id

    # Fitting tokenizer for selected model
    tokenizer = AutoTokenizer.from_pretrained(args.model, cache_dir=args.cache)

    # Fitting model for token classification
    model = AutoModelForTokenClassification.from_pretrained(
        args.model, cache_dir=args.cache, config=config,
    )

    training_args = TrainingArguments(
        output_dir=args.output,
        learning_rate=2e-5,
        per_device_train_batch_size=16,
        per_device_eval_batch_size=16,
        num_train_epochs=2,
        weight_decay=0.01,
        eval_strategy="epoch",
        save_strategy="epoch",
        logging_strategy="steps",
        logging_steps=5,
        report_to="tensorboard",
        logging_dir=args.logs,
    )

    data_collator = DataCollatorForTokenClassification(tokenizer)

    with open(args.data, 'r', encoding='utf-8') as f:
        dataset = json.load(f)

    trainer = Trainer(
        model=model,
        args=training_args,
        train_dataset=dataset['train'],
        eval_dataset=dataset['validation'],
        processing_class=tokenizer,
        data_collator=data_collator,
    )

    print('Training...')
    trainer.train()

    print('Saving...')
    model.save_pretrained(args.output)



# All valid POS tags
pos_tags = [
    'NNG', 'NNP', 'NNB',
    'NR', 'NP',
    'VV', 'VA', 'VX', 'VCP', 'VCN',
    'MM', 'MAG', 'MAJ', 'IC',
    'JKS', 'JKC', 'JKG', 'JKO', 'JKB', 'JKV', 'JKQ', 'JX', 'JC',
    'EP', 'EF', 'EC', 'ETN', 'ETM',
    'XPN', 'XSN', 'XSV', 'XSA', 'XR',
    'SF', 'SP', 'SS', 'SE', 'SO', 'SW', 'SH', 'SL', 'SN',
]

# All valid labels
#  O: Other/Nothing
#  B-POS: Begin POS
#  I-POS: Inside POS (continuation of B-POS tag)
all_labels = ['O'] + [f'B-{tag}' for tag in pos_tags] + [f'I-{tag}' for tag in pos_tags]

label2id = {label: i for i, label in enumerate(all_labels)}
id2label = {i: label for i, label in enumerate(all_labels)}
num_labels = len(all_labels)



import os
import json
import random
import unicodedata
from tqdm import tqdm
from transformers import AutoTokenizer, AutoConfig
import kiwipiepy

from .labels import pos_tags, label2id


kiwi = kiwipiepy.Kiwi(num_workers=0, model_type='sbg')

def kiwi_tag_map(tag):
    r = {
        'SSO': 'SS',
        'SSC': 'SS',
        'XSM': 'XSA',
        'SB': None,
        'W_URL': None,
        'W_EMAIL': None,
        'W_HASHTAG': None,
        'W_MENTION': None,
        'W_SERIAL': None,
        'W_EMOJI': None,
        'Z_CODA': None,
        'Z_SIOT': None,
        'USER0': None,
        'USER1': None,
        'USER2': None,
        'USER3': None,
        'USER4': None,
    }.get(tag, tag)

    if tag.endswith('-I'):
        r = r[:-2]
    if tag.endswith('-R'):
        r = r[:-2]

    if r is None:
        raise ValueError(f'Unknown POS: {tag}')
    
    return r

def process_sents(tokenizer, normalize_func, sents, max_tokens):
    """
    Assigns POS tags emitted by Kiwi to tokens in the sentence.
    """
    entries = []

    for sent in sents:
        text = sent.text
        kiwi_tokens = sent.tokens

        try:
            morphs = [
                {
                    'form': kiwi_token.form,
                    'tag': kiwi_tag_map(kiwi_token.tag),
                    'start': kiwi_token.start - sent.start,
                    'end': kiwi_token.start + kiwi_token.len - sent.start,
                }
                for kiwi_token in kiwi_tokens
            ]
        except ValueError:
            # ignore sentences with unsupported POS morphs
            continue

        morph_ends = set()
        for morph in morphs:
            morph_ends.add(morph['end'])
        
        # Normalize text and create a mapping from normalized text to original text
        og_map = {}
        text_norm = ''

        for text_idx, text_char in enumerate(text):
            og_map[len(text_norm)] = text_idx
            text_norm += normalize_func(text_char)
        og_map[len(text_norm)] = len(text)
        
        last_idx = 0
        for i in range(len(text_norm) + 1):
            if i in og_map:
                last_idx = og_map[i]
            else:
                og_map[i] = last_idx
        
        tokenized_result = tokenizer(text_norm, return_offsets_mapping=True)
        tokens = tokenizer.convert_ids_to_tokens(tokenized_result['input_ids'])

        # Skip sentences surpassing the token limit of the architecture
        if max_tokens > 0 and len(tokens) > max_tokens:
            # print(f'Skipping long sentence with {len(tokens)} tokens: {text[:50]}...')
            continue
        
        labels = []
        is_after_space_list = []

        # Assign labels to tokens based on Kiwi Tokens
        for id, (start, end), token in zip(tokenized_result['input_ids'], tokenized_result['offset_mapping'], tokens):
            og_start = og_map[start]
            og_end = og_map[end]

            is_empty_token = token in tokenizer.all_special_tokens or start == end
            is_after_split = og_start in morph_ends or is_empty_token or start == 0

            possible_pos = set()
            # find all pos tags that overlap with the token
            for morph in morphs:
                if morph['start'] < og_end and morph['end'] > og_start:
                    possible_pos.add(morph['tag'])
            if len(possible_pos) < 1:
                pos = None
            elif len(possible_pos) == 1:
                pos = possible_pos.pop()
            if len(possible_pos) > 1:
                pos = min(possible_pos, key=lambda tag: pos_tags.index(tag))

            if pos is None:
                label = 'O'
            else:
                if is_after_split:
                    is_continue = False
                else:
                    is_continue = len(labels) > 0 and labels[-1] == f'B-{pos}'
                label = f'I-{pos}' if is_continue else f'B-{pos}'

            labels.append(label)
            
            prev_char = text[og_start - 1] if og_start > 0 else ' '
            is_after_space_list.append(prev_char.isspace())

        if labels.count('O') - 2 > len(labels) * 0.1:
            print(f'Skipping sentence with too many O labels: {text[:50]}...')
            continue

        entries.append({
            'input_ids': [id for id in tokenized_result['input_ids']],
            'attention_mask': [1] * len(tokenized_result['input_ids']),
            'labels': [label2id[label] for label in labels],
            # 'is_after_space': is_after_space_list,
        })

    return entries

def process_lines(tokenizer, normalize_func, lines, max_tokens):
    multi_sents_txt = [
        unicodedata.normalize('NFC', line)
        for line in lines
        if line.strip()
    ]
    kiwi_results = kiwi.split_into_sents(multi_sents_txt, return_tokens=True)

    entries = []
    for kiwi_result in kiwi_results:
        entries.extend(
            process_sents(tokenizer, normalize_func, kiwi_result, max_tokens)
        )
    return entries

def chunked(iterable, n):
    chunk = []
    for item in iterable:
        chunk.append(item)
        if len(chunk) >= n:
            yield chunk
            chunk = []
    if chunk:
        yield chunk

def data(
    model,
    cache,
    input,
    normalize_mode,
    output,
    split,
    **_kwargs,
):
    tokenizer = AutoTokenizer.from_pretrained(model, cache_dir=cache)
    config = AutoConfig.from_pretrained(model)
    max_token_length = config.max_position_embeddings

    txt_files = []
    if os.path.isfile(input):
        txt_files.append(input)
    elif os.path.isdir(input):
        for root, _, files in os.walk(input):
            for file in files:
                if file.endswith('.txt'):
                    txt_files.append(os.path.join(root, file))
    else:
        raise ValueError(f'Invalid input: {input}')
    txt_files.sort()
    txt_files_iter = tqdm(txt_files) if len(txt_files) > 1 else txt_files

    data = []

    if normalize_mode is None:
        normalize_func = lambda x: x
    else:
        normalize_func = lambda x: unicodedata.normalize(normalize_mode, x)

    for txt_file in txt_files_iter:
        total = 0
        with open(txt_file, 'r', encoding='utf-8') as f:
            for _line in f:
                total += 1
        with open(txt_file, 'r', encoding='utf-8') as f:
            for lines in chunked(tqdm(f, leave=False, desc=txt_file, total=total), 500):
                data.extend(
                    process_lines(tokenizer, normalize_func, lines, max_token_length)
                )

    print('Shuffling data...')
    random.shuffle(data)

    print('Splitting data...')
    train_size = int(len(data) * split)
    train_data = data[:train_size]
    validation_data = data[train_size:]

    data = {
        'train': train_data,
        'validation': validation_data,
    }

    if output is None:
        return data

    print(f'Writing data to {output}...',)
    with open(output, 'w', encoding='utf-8') as f:
        json.dump(data, f)



from setuptools import setup

def read_requirements():
    with open('requirements.txt') as f:
        return [line.strip() for line in f if line.strip() and not line.startswith('#')]

setup(
    name='kotok',
    version='1.0.0',
    packages=['kotok'],
    install_requires=read_requirements(),
    author='Daeun Jung',
    author_email='Daeun.Jung@ruhr-uni-bochum.de',
    description='Korean morphological analyzer based on the BERT architecture',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    url='http://github.com/Daeun271/kotok',
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Operating System :: OS Independent',
    ],
    python_requires='>=3.12',
)



import os
import logging
import argparse

def make_parser():
    """
    Construct the command line parser for the CLI.
    """
    parser = argparse.ArgumentParser()

    parser.add_argument('-v', '--verbose', action='store_true')

    subparsers = parser.add_subparsers(dest='command')

    model_default = 'klue/bert-base'
    classification_model_default = os.path.join('models', 'kotok_model')
    error_classification_model_default = os.path.join('models', 'kotok_error_model')
    spacing_classification_model_default = os.path.join('models', 'kotok_spacing_model')
    cache_default = os.path.join('cache')
    data_default = os.path.join('data', 'labeled.json')
    lemma_data_default = os.path.join('data', 'lemma')

    data_dl = subparsers.add_parser('data_dl')
    data_dl.add_argument('-o', '--out_dir', type=str, default=os.path.join('data', 'txt'), help='Output directory for the downloaded data')

    data = subparsers.add_parser('data')
    data.add_argument('-m', '--model', type=str, default=model_default, help='Pretrained model name or path for tokenization')
    data.add_argument('-c', '--cache', type=str, default=cache_default, help='Cache directory')
    data.add_argument('-i', '--input', type=str, default=os.path.join('data', 'txt'), help='Input data directory containing text files')
    data.add_argument('-n', '--normalize_mode', type=str, default=None, help='Unicode normalization mode')
    data.add_argument('-o', '--output', type=str, default=data_default, help='Output file path')
    data.add_argument('-s', '--split', type=float, default=0.8, help='Train-test split ratio')

    train = subparsers.add_parser('train')
    train.add_argument('-m', '--model', type=str, default=model_default, help='Pretrained model name or path for tokenization')
    train.add_argument('-c', '--cache', type=str, default=cache_default, help='Cache directory')
    train.add_argument('-d', '--data', type=str, default=data_default, help='Data file path, generated by the data command')
    train.add_argument('-o','--output', type=str, default=classification_model_default, help='Output directory for the trained model')
    train.add_argument('-l', '--logs', type=str, default='logs', help='Output directory for the logs')

    inference = subparsers.add_parser('inference')
    inference.add_argument('-cm', '--classification_model', type=str, default=classification_model_default, help='Classification model path, generated by the train command')
    inference.add_argument('-m', '--model', type=str, default=model_default, help='Pretrained model name or path for tokenization')
    inference.add_argument('-c', '--cache', type=str, default=cache_default, help='Cache directory')
    inference.add_argument('-f', '--format', type=str, default='pretty', help='Output format')
    inference.add_argument('-n', '--normalize_mode', type=str, default=None, help='Unicode normalization mode')
    inference.add_argument('-u', '--user_dict', type=str, default=None, help='User dictionary file or directory path')
    inference.add_argument('-ld', '--lemma_data', type=str, default=lemma_data_default, help='Lemmatization data directory')
    inference.add_argument('-nl', '--no_lemma', action='store_true', default=False, help='Disable lemmatization')
    inference.add_argument('-ne', '--no_error_correction', action='store_true', default=False, help='Disable error correction')
    inference.add_argument('-ecm', '--error_classification_model', type=str, default=error_classification_model_default, help='Error classification model path, generated by the train command')
    inference.add_argument('-em', '--error_model', type=str, default=model_default, help='Pretrained model name or path for error correction')
    inference.add_argument('-ns', '--no_spacing_correction', action='store_true', default=False, help='Disable spacing correction')
    inference.add_argument('-scm', '--spacing_classification_model', type=str, default=spacing_classification_model_default, help='Spacing classification model path, generated by the train command')
    inference.add_argument('-sm', '--spacing_model', type=str, default=model_default, help='Pretrained model name or path for spacing correction')

    lemmatize = subparsers.add_parser('lemmatize')
    lemmatize.add_argument('-d', '--data-dir', type=str, default=lemma_data_default, help='Lemmatization data directory')

    return parser

def main():
    logging.basicConfig(level=logging.INFO, format='[%(levelname)s] %(message)s')

    args = make_parser().parse_args()

    if args.verbose:
        logging.getLogger().setLevel(logging.DEBUG)

    # Only import sub modules if needed to reduce startup time
    if args.command == 'data_dl':
        from .data_dl import data_dl
        data_dl(**args.__dict__)
    elif args.command == 'data':
        from .data import data
        data(**args.__dict__)
    elif args.command == 'train':
        from .train import train
        train(args)
    elif args.command == 'inference':
        from .inference import inference
        # If no error or spacing correction is needed, set the model to None to override the default model
        if args.no_error_correction:
            args.error_model = None
            args.error_classification_model = None
        if args.no_spacing_correction:
            args.spacing_model = None
            args.spacing_classification_model = None    
        inference(**args.__dict__)
    elif args.command == 'lemmatize':
        from .lemmatize import lemmatize
        lemmatize(**args.__dict__)


if __name__ == '__main__':
    main()



from transformers import pipeline, AutoTokenizer
import dataclasses
import unicodedata
import os
import numpy as np
import logging
from .lemmatize import Lemmatizer

@dataclasses.dataclass
class UserDictEntry:
    morph: str
    suffix_wildcard: bool
    pos: str
    pos_match: set[str] | None

@dataclasses.dataclass
class Token:
    surface: str
    lemma: str
    tag: str
    start: int
    end: int

    def __repr__(self):
        if self.surface == self.lemma:
            return f'{self.surface}/{self.tag}'
        return f'{self.surface}/{self.lemma}/{self.tag}'
    
    def __str__(self):
        return repr(self)


def normalize_with_map(text, normalize_mode):
    convert_map = {}
    text_norm = ''

    for text_idx, text_char in enumerate(text):
        convert_map[len(text_norm)] = text_idx
        text_norm += unicodedata.normalize(normalize_mode, text_char)
    convert_map[len(text_norm)] = len(text)
    
    last_idx = 0
    for i in range(len(text_norm) + 1):
        if i in convert_map:
            last_idx = convert_map[i]
        else:
            convert_map[i] = last_idx

    return text_norm, convert_map


def apply_splits_single(token: Token):
    r = []

    space_idx = token.surface.find(' ')
    if space_idx == -1:
        return [token]
    
    space_idx_end = space_idx
    while token.surface[space_idx_end].isspace():
        space_idx_end += 1
    
    logging.debug(f'Splitting: "{token.surface}" -> "{token.surface[:space_idx]}" / "{token.surface[space_idx_end:]}"')

    return [
        Token(
            surface = token.surface[:space_idx],
            lemma = token.surface[:space_idx],
            tag = token.tag,
            start = token.start,
            end = token.start + space_idx,
        ),
    ] + apply_splits_single(
        Token(
            surface = token.surface[space_idx_end:],
            lemma = token.surface[space_idx_end:],
            tag = token.tag,
            start = token.start + space_idx_end,
            end = token.end,
        )
    )

def apply_splits(tokens: list[Token]):
    r = []
    for token in tokens:
        r += apply_splits_single(token)
    return r


def apply_lemmatization(tokens: list[Token], lemmatizer: Lemmatizer):
    r = []
    for token in tokens:
        if not token.tag.startswith('V'):
            r.append(token)
            continue
        lemma, extra_tokens = lemmatizer.lemmatize(token.surface)
        token.lemma = lemma
        token.surface = token.surface[:len(lemma)]
        token.end = token.start + len(lemma)
        for extra_token in extra_tokens:
            extra_token.start += token.start
            extra_token.end += token.start
        r.append(token)
        r += extra_tokens
    return r


def analyze(
    classification_pipeline,
    text,
    normalize_mode=None,
    lemmatizer=None
):
    text_norm, convert_map = normalize_with_map(text, normalize_mode or 'NFC')

    raw_tokens = classification_pipeline(text_norm)

    i = 0
    tokens = []

    while i < len(raw_tokens):
        raw_token = raw_tokens[i]
        entity = raw_token['entity']

        end_i = i

        pos = f'<UNK-{entity}>'
        if entity.startswith('B-'):
            pos = entity[2:]

            for j in range(i + 1, len(raw_tokens)):
                if raw_tokens[j]['entity'] == f'I-{pos}':
                    end_i = j
                else:
                    break
        
        start = raw_token['start']
        end = raw_tokens[end_i]['end']

        og_start = convert_map[start]
        og_end = convert_map[end]

        surface = text[og_start:og_end]

        token = Token(
            surface = surface,
            lemma = surface,
            tag = pos,
            start = og_start,
            end = og_end,
        )
        tokens.append(token)

        i = end_i + 1

    tokens = apply_splits(tokens)
    if lemmatizer:
        tokens = apply_lemmatization(tokens, lemmatizer)
    
    return tokens


# TODO: Support normalize mode, then merge with analyze
def analyze_with_user_dict(
    classification_pipeline,
    text,
    normalize_mode=None,
    user_dict=[],
    lemmatizer=None,
    _ignore_user_dict_entries=[]
):

    mask_token = classification_pipeline.tokenizer.mask_token

    tokens_pre_masked = classification_pipeline(text)
    
    logging.debug('pre-masked')
    for token in tokens_pre_masked:
        logging.debug(token)

    pre_mask_char = '\x1A'
    pre_masks = []
    text_pre_masked = text

    # mask text that appears in the user dictionary
    for entry_idx, entry in enumerate(user_dict):
        search_idx = 0
        while True:
            idx = text_pre_masked.find(entry.morph, search_idx)
            if idx == -1:
                break

            pre_mask_start = idx
            pre_mask_end = idx + len(entry.morph)

            if entry.suffix_wildcard:
                for token in tokens_pre_masked:
                    if token['start'] <= pre_mask_start and pre_mask_end <= token['end']:
                        pre_mask_end = max(token['end'], pre_mask_end)

            '''
            # These are now handled post-masking, pos tags for user dict entries are not predicted properly in many cases
            if entry.pos_match:
                found_pos = False
                for token in tokens_pre_masked:
                    pos = token['entity'].split('-')[-1]
                    if pre_mask_start <= token['start'] and token['end'] <= pre_mask_end and pos in entry.pos:
                        found_pos = True
                        break
                if not found_pos:
                    search_idx = pre_mask_end
                    continue
            '''

            replace_len = pre_mask_end - pre_mask_start

            search_idx = pre_mask_end + 1

            new_entry = (pre_mask_start, pre_mask_end, entry_idx)
            if new_entry in _ignore_user_dict_entries:
                continue

            text_pre_masked = text_pre_masked[:pre_mask_start] + pre_mask_char * replace_len + text_pre_masked[pre_mask_end:]
            pre_masks.append(new_entry)

    pre_masks_ftb = sorted(pre_masks, key=lambda x: x[0])
    pre_masks_btf = sorted(pre_masks, key=lambda x: x[0], reverse=True)

    text_masked = text_pre_masked
    for pre_mask_start, pre_mask_end, entry_idx in pre_masks_btf:
        text_masked = text_masked[:pre_mask_start] + mask_token + text_masked[pre_mask_end:]

    tokens = classification_pipeline(text_masked)
    logging.debug('masked')
    for token in tokens:
        logging.debug(token)

    mask_idx = 0
    offset = 0

    ignore_user_dict_entries = []

    # Check masked tokens
    # If the predicted pos does not match the user dict entry, ignore the entry
    for token in tokens:
        if token['word'] == mask_token:
            entry = pre_masks_ftb[mask_idx]
            user_entry = user_dict[entry[2]]

            if user_entry.pos_match:
                predicted_pos = token['entity'].split('-')[-1]
                if not predicted_pos in user_entry.pos_match:
                    ignore_user_dict_entries.append(entry)
                    mask_idx += 1
                    continue

            offset = entry[1] - token['end']
            
            token['score'] = np.float32(1.0)
            token['word'] = user_entry.morph
            token['tag'] = f'B-{user_entry.pos}'
            token['start'] = entry[0]
            token['end'] = entry[1]

            mask_idx += 1
            continue

        token['start'] = token['start'] + offset
        token['end'] = token['end'] + offset

    assert mask_idx == len(pre_masks_ftb)

    if ignore_user_dict_entries:
        # retry, with the ignored entries
        return analyze_with_user_dict(classification_pipeline, text, normalize_mode, user_dict, ignore_user_dict_entries)

    logging.debug('fixed')
    for token in tokens:
        logging.debug(token)

    i = 0
    tokens_out = []

    # Combine B- and I- tokens
    while i < len(tokens):
        raw_token = tokens[i]
        entity = raw_token['entity']

        end_i = i

        pos = f'<UNK-{entity}>'
        if entity.startswith('B-'):
            pos = entity[2:]

            for j in range(i + 1, len(tokens)):
                if tokens[j]['entity'] == f'I-{pos}':
                    end_i = j
                else:
                    break
        
        start = raw_token['start']
        end = tokens[end_i]['end']
        surface = text[start:end]

        token_out = Token(
            surface = surface,
            lemma = surface,
            tag = pos,
            start = start,
            end = end,
        )
        tokens_out.append(token_out)

        i = end_i + 1
    
    if not _ignore_user_dict_entries:
        # prevent morphemes across spaces
        tokens_out = apply_splits(tokens_out)

        # lemmatize verbs
        if lemmatizer:
            tokens_out = apply_lemmatization(tokens_out, lemmatizer)

    return tokens_out


def create_pipeline(
    model,
    classification_model,
    cache,
):
    tokenizer = AutoTokenizer.from_pretrained(model, cache_dir=cache)

    return pipeline(
        'token-classification',
        model=classification_model,
        tokenizer=tokenizer,
        ignore_labels=[],
    )


def inference(
    model,
    classification_model,
    cache,
    format,
    normalize_mode = None,
    user_dict = None,
    lemma_data = None,
    no_lemma = False,
    error_model = None,
    error_classification_model = None,
    spacing_model = None,
    spacing_classification_model = None,
    **kwargs,
):
    if isinstance(user_dict, str):
        user_dict = load_user_dict(user_dict)
    elif user_dict is None:
        user_dict = []
    elif not isinstance(user_dict, list):
        raise ValueError(f'Invalid user dictionary: {user_dict}')
    
    lemmatizer = None
    if not no_lemma or lemma_data:
        lemmatizer = Lemmatizer(lemma_data)

    spacing_corrector = None
    if spacing_model and spacing_classification_model:
        from .spacing.inference import correct as correct_spacing, create_pipeline as create_spacing_pipeline
        spacing_pipeline = create_spacing_pipeline(
            spacing_model,
            spacing_classification_model,
            cache,
        )
        spacing_corrector = lambda text: correct_spacing(spacing_pipeline, text)

    error_corretor = None
    if error_model and error_classification_model:
        from .error.inference import create_error_corrector
        error_corretor = create_error_corrector(error_model, error_classification_model, cache)

    classification_pipeline = create_pipeline(
        model,
        classification_model,
        cache,
    )

    if user_dict:
        analyze_func = lambda text: analyze_with_user_dict(classification_pipeline, text, normalize_mode, user_dict, lemmatizer)
    else:
        analyze_func = lambda text: analyze(classification_pipeline, text, normalize_mode, lemmatizer)

    while True:
        try:
            text = input('> ')

            if spacing_corrector:
                text = spacing_corrector(text)

            if error_corretor:
                text, _corrections = error_corretor(text)

            if normalize_mode:
                text = unicodedata.normalize(normalize_mode, text)
            if format == 'pretty':
                tokens = analyze_func(text)
                print(' '.join(map(str, tokens)))
            elif format == 'raw':
                tokens_raw = classification_pipeline(text)
                for token in tokens_raw:
                    print(token)
        except KeyboardInterrupt:
            break
        except EOFError:
            break


class Analyzer:
    def __init__(
        self,
        model: str,
        classification_model: str,
        cache: str | None = None,
        normalize_mode: list[UserDictEntry] | str | None = None,
        user_dict: str | None = None,
        lemma_data: str | None = None,
        no_lemma: bool = False,
        error_model: str | None = None,
        error_classification_model: str | None = None,
        spacing_model: str | None = None,
        spacing_classification_model: str | None = None,
        **kwargs,
    ):
        """
        model: str -- The tokenizer model to use, either a name on Hugging Face or a path to a local model
        classification_model: str -- The classification model to use, generated from the train command
        cache: str | None -- The cache directory to use for the tokenizer
        normalize_mode: str | None -- The unicode normalization mode to use for the input text
        user_dict: list[UserDictEntry] | str | None -- The user dictionary to use for the analyzer, either a list of UserDictEntry objects, a path to a file, or a path to a directory
        lemma_data: str | None -- Path to the lemmatization data directory
        no_lemma: bool -- Whether to force disable lemmatization
        error_model: str | None -- The tokenizer model to use for the error corrector, either a name on Hugging Face or a path to a local model
        error_classification_model: str | None -- The classification model to use for the error corrector, generated from the train command
        spacing_model: str | None -- The tokenizer model to use for the spacing corrector, either a name on Hugging Face or a path to a local model
        spacing_classification_model: str | None -- The classification model to use for the spacing corrector, generated from the train command
        """

        self.normalize_mode = normalize_mode

        if isinstance(user_dict, str):
            self.user_dict = load_user_dict(user_dict)
        elif user_dict is None:
            self.user_dict = []
        elif not isinstance(user_dict, list):
            raise ValueError(f'Invalid user dictionary: {user_dict}')
        
        self.lemmatizer = None
        if not no_lemma or lemma_data:
            self.lemmatizer = Lemmatizer(lemma_data)

        self.spacing_corrector = None
        if spacing_model and spacing_classification_model:
            from .spacing.inference import correct as correct_spacing, create_pipeline as create_spacing_pipeline
            spacing_pipeline = create_spacing_pipeline(
                spacing_model,
                spacing_classification_model,
                cache,
            )
            self.spacing_corrector = lambda text: correct_spacing(spacing_pipeline, text)

        self.error_corretor = None
        if error_model and error_classification_model:
            from .error.inference import create_error_corrector
            self.error_corretor = create_error_corrector(error_model, error_classification_model, cache)

        self.classification_pipeline = create_pipeline(
            model,
            classification_model,
            cache,
        )

        if self.user_dict:
            self.analyze_func = lambda text: analyze_with_user_dict(self.classification_pipeline, text, self.normalize_mode, self.user_dict, self.lemmatizer)
        else:
            self.analyze_func = lambda text: analyze(self.classification_pipeline, text, self.normalize_mode, self.lemmatizer)

    def run(self, text: str, format='pretty') -> list[Token]:
        """
        text: str -- The input text to analyze
        format: str -- The output format, either 'pretty' or 'raw'. 'pretty' will return a list of Token objects, 'raw' will return the raw output from the model.
        """

        if self.spacing_corrector:
            text = self.spacing_corrector(text)

        if self.error_corretor:
            text, _corrections = self.error_corretor(text)

        if self.normalize_mode:
            text = unicodedata.normalize(self.normalize_mode, text)

        if format == 'raw':
            return self.classification_pipeline(text)

        return self.analyze_func(text)        


#
# User dictionary loading
#

def load_user_dict_file(file_path: str, _no_sort: bool=False):
    '''
    file_path: str -- 
    '''

    user_dict = []

    with open(file_path, 'r', encoding='utf-8') as file:
        for line_idx, line in enumerate(file):
            line = line.strip()
            if not line or line.startswith('#'):
                continue

            line = line.replace('\t', ' ')
            space_idx = line.rfind(' ')
            if space_idx == -1:
                raise ValueError(f'Invalid user dictionary entry at {file_path}:{line_idx + 1}')
            
            suffix_wildcard = False
            pos_match = None
            morph = line[:space_idx].strip()
            pos_raw = line[space_idx + 1:].strip()

            if morph.endswith('*'):
                morph = morph[:-1].strip()
                suffix_wildcard = True

            exclamation_idx = pos_raw.rfind('!')
            if exclamation_idx != -1:
                pos = pos_raw[:exclamation_idx].strip()
                pos_parts = pos_raw[exclamation_idx + 1:].split(',')
                if pos_parts:
                    pos_match = set(pos_parts)
                else:
                    pos_match = set([pos])
            else:
                pos = pos_raw

            user_dict.append(UserDictEntry(
                morph = morph,
                suffix_wildcard = suffix_wildcard,
                pos = pos,
                pos_match = pos_match,
            ))

    if not _no_sort:
        user_dict.sort(key=lambda x: len(x.morph), reverse=True)

    return user_dict

def load_user_dict_dir(dir_path: str, _no_sort: bool=True) -> list[UserDictEntry]:
    user_dict = []
    for root, _, files in os.walk(dir_path):
        for file in files:
            if os.path.splitext(file)[1] in ('.tsv'):
                file_path = os.path.join(root, file)
                user_dict.extend(load_user_dict_file(file_path, _no_sort=True))
    if not _no_sort:
        user_dict = dict(sorted(user_dict.items(), key=lambda x: len(x[0]), reverse=True))
    return user_dict

def load_user_dict(path: str) -> list[UserDictEntry]:
    if os.path.isfile(path):
        return load_user_dict_file(path)
    elif os.path.isdir(path):
        return load_user_dict_dir(path)
    else:
        raise ValueError(f'Invalid user dictionary: {path}')
